==> bodo-2022.3-cp38-cp38-macosx_10_15_x86_64.whl: bodo/libs/binary_arr_ext.py <==
"""Array implementation for binary (bytes) objects, which are usually immutable.
It is equivalent to string array, except that it stores a 'bytes' object for each
element instead of 'str'.
"""
import operator
import llvmlite.binding as ll
import numba
import numpy as np
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import intrinsic, lower_builtin, lower_cast, make_attribute_wrapper, overload, overload_attribute, overload_method
import bodo
from bodo.libs import hstr_ext
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.utils.typing import BodoError, is_list_like_index_type
_bytes_fromhex = types.ExternalFunction('bytes_fromhex', types.int64(types.
    voidptr, types.voidptr, types.uint64))
ll.add_symbol('bytes_to_hex', hstr_ext.bytes_to_hex)
ll.add_symbol('bytes_fromhex', hstr_ext.bytes_fromhex)
bytes_type = types.Bytes(types.uint8, 1, 'C', readonly=True)


class BinaryArrayType(types.IterableType, types.ArrayCompatible):

    def __init__(self):
        super(BinaryArrayType, self).__init__(name='BinaryArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return bytes_type

    def copy(self):
        return BinaryArrayType()

    @property
    def iterator_type(self):
        return bodo.utils.typing.BodoArrayIterator(self)


binary_array_type = BinaryArrayType()


@overload(len, no_unliteral=True)
def bin_arr_len_overload(bin_arr):
    if bin_arr == binary_array_type:
        return lambda bin_arr: len(bin_arr._data)


@overload_attribute(BinaryArrayType, 'size')
def bin_arr_size_overload(bin_arr):
    return lambda bin_arr: len(bin_arr._data)


@overload_attribute(BinaryArrayType, 'shape')
def bin_arr_shape_overload(bin_arr):
    return lambda bin_arr: (len(bin_arr._data),)


@overload_attribute(BinaryArrayType, 'nbytes')
def bin_arr_nbytes_overload(bin_arr):
    return lambda bin_arr: bin_arr._data.nbytes


@overload_attribute(BinaryArrayType, 'ndim')
def overload_bin_arr_ndim(A):
    return lambda A: 1


@overload_attribute(BinaryArrayType, 'dtype')
def overload_bool_arr_dtype(A):
    return lambda A: np.dtype('O')


@numba.njit
def pre_alloc_binary_array(n_bytestrs, n_chars):
    if n_chars is None:
        n_chars = -1
    bin_arr = init_binary_arr(bodo.libs.array_item_arr_ext.
        pre_alloc_array_item_array(np.int64(n_bytestrs), (np.int64(n_chars)
        ,), bodo.libs.str_arr_ext.char_arr_type))
    if n_chars == 0:
        bodo.libs.str_arr_ext.set_all_offsets_to_0(bin_arr)
    return bin_arr


@intrinsic
def init_binary_arr(typingctx, data_typ=None):
    assert isinstance(data_typ, ArrayItemArrayType
        ) and data_typ.dtype == types.Array(types.uint8, 1, 'C')

    def codegen(context, builder, sig, args):
        clek__ojdi, = args
        itee__efocg = context.make_helper(builder, binary_array_type)
        itee__efocg.data = clek__ojdi
        context.nrt.incref(builder, data_typ, clek__ojdi)
        return itee__efocg._getvalue()
    return binary_array_type(data_typ), codegen


@intrinsic
def init_bytes_type(typingctx, data_typ, length_type):
    assert data_typ == types.Array(types.uint8, 1, 'C')
    assert length_type == types.int64

    def codegen(context, builder, sig, args):
        nrufx__ilw = cgutils.create_struct_proxy(sig.args[0])(context,
            builder, value=args[0])
        xqly__vnie = args[1]
        ijxt__wzm = cgutils.create_struct_proxy(bytes_type)(context, builder)
        ijxt__wzm.meminfo = context.nrt.meminfo_alloc(builder, xqly__vnie)
        ijxt__wzm.nitems = xqly__vnie
        ijxt__wzm.itemsize = lir.Constant(ijxt__wzm.itemsize.type, 1)
        ijxt__wzm.data = context.nrt.meminfo_data(builder, ijxt__wzm.meminfo)
        ijxt__wzm.parent = cgutils.get_null_value(ijxt__wzm.parent.type)
        ijxt__wzm.shape = cgutils.pack_array(builder, [xqly__vnie], context
            .get_value_type(types.intp))
        ijxt__wzm.strides = nrufx__ilw.strides
        cgutils.memcpy(builder, ijxt__wzm.data, nrufx__ilw.data, xqly__vnie)
        return ijxt__wzm._getvalue()
    return bytes_type(data_typ, length_type), codegen


@intrinsic
def cast_bytes_uint8array(typingctx, data_typ):
    assert data_typ == bytes_type

    def codegen(context, builder, sig, args):
        return impl_ret_borrowed(context, builder, sig.return_type, args[0])
    return types.Array(types.uint8, 1, 'C')(data_typ), codegen


@overload_method(BinaryArrayType, 'copy', no_unliteral=True)
def binary_arr_copy_overload(arr):

    def copy_impl(arr):
        return init_binary_arr(arr._data.copy())
    return copy_impl


@overload_method(types.Bytes, 'hex')
def binary_arr_hex(arr):
    blof__jdkr = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

    def impl(arr):
        xqly__vnie = len(arr) * 2
        output = numba.cpython.unicode._empty_string(blof__jdkr, xqly__vnie, 1)
        bytes_to_hex(output, arr)
        return output
    return impl


@lower_cast(types.CPointer(types.uint8), types.voidptr)
def cast_uint8_array_to_voidptr(context, builder, fromty, toty, val):
    return val


make_attribute_wrapper(types.Bytes, 'data', '_data')


@overload_method(types.Bytes, '__hash__')
def bytes_hash(arr):

    def impl(arr):
        return numba.cpython.hashing._Py_HashBytes(arr._data, len(arr))
    return impl


@intrinsic
def bytes_to_hex(typingctx, output, arr):

    def codegen(context, builder, sig, args):
        yinfz__deniy = cgutils.create_struct_proxy(sig.args[0])(context,
            builder, value=args[0])
        jzwbd__ebmm = cgutils.create_struct_proxy(sig.args[1])(context,
            builder, value=args[1])
        ajvet__rrpp = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(64)])
        qjzfo__wqv = cgutils.get_or_insert_function(builder.module,
            ajvet__rrpp, name='bytes_to_hex')
        builder.call(qjzfo__wqv, (yinfz__deniy.data, jzwbd__ebmm.data,
            jzwbd__ebmm.nitems))
    return types.void(output, arr), codegen


@overload(operator.getitem, no_unliteral=True)
def binary_arr_getitem(arr, ind):
    if arr != binary_array_type:
        return
    if isinstance(ind, types.Integer):

        def impl(arr, ind):
            kgh__ycltj = arr._data[ind]
            return init_bytes_type(kgh__ycltj, len(kgh__ycltj))
        return impl
    if is_list_like_index_type(ind) and (ind.dtype == types.bool_ or
        isinstance(ind.dtype, types.Integer)) or isinstance(ind, types.
        SliceType):
        return lambda arr, ind: init_binary_arr(arr._data[ind])
    raise BodoError(
        f'getitem for Binary Array with indexing type {ind} not supported.')


def bytes_fromhex(hex_str):
    pass


@overload(bytes_fromhex)
def overload_bytes_fromhex(hex_str):
    hex_str = types.unliteral(hex_str)
    if hex_str == bodo.string_type:
        blof__jdkr = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

        def impl(hex_str):
            if not hex_str._is_ascii or hex_str._kind != blof__jdkr:
                raise TypeError(
                    'bytes.fromhex is only supported on ascii strings')
            clek__ojdi = np.empty(len(hex_str) // 2, np.uint8)
            xqly__vnie = _bytes_fromhex(clek__ojdi.ctypes, hex_str._data,
                len(hex_str))
            qegj__xtvfs = init_bytes_type(clek__ojdi, xqly__vnie)
            return qegj__xtvfs
        return impl
    raise BodoError(f'bytes.fromhex not supported with argument type {hex_str}'
        )


@overload(operator.setitem)
def binary_arr_setitem(arr, ind, val):
    if arr != binary_array_type:
        return
    if val == types.none or isinstance(val, types.optional):
        return
    if val != bytes_type:
        raise BodoError(
            f'setitem for Binary Array only supported with bytes value and integer indexing'
            )
    if isinstance(ind, types.Integer):

        def impl(arr, ind, val):
            arr._data[ind] = bodo.libs.binary_arr_ext.cast_bytes_uint8array(val
                )
        return impl
    raise BodoError(
        f'setitem for Binary Array with indexing type {ind} not supported.')


def create_binary_cmp_op_overload(op):

    def overload_binary_cmp(lhs, rhs):
        myhqa__cscb = lhs == binary_array_type
        hvg__iaa = rhs == binary_array_type
        qzw__pkjiw = 'lhs' if myhqa__cscb else 'rhs'
        cgyx__glqje = 'def impl(lhs, rhs):\n'
        cgyx__glqje += '  numba.parfors.parfor.init_prange()\n'
        cgyx__glqje += f'  n = len({qzw__pkjiw})\n'
        cgyx__glqje += (
            '  out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(n)\n')
        cgyx__glqje += '  for i in numba.parfors.parfor.internal_prange(n):\n'
        qmuz__ukmr = []
        if myhqa__cscb:
            qmuz__ukmr.append('bodo.libs.array_kernels.isna(lhs, i)')
        if hvg__iaa:
            qmuz__ukmr.append('bodo.libs.array_kernels.isna(rhs, i)')
        cgyx__glqje += f"    if {' or '.join(qmuz__ukmr)}:\n"
        cgyx__glqje += '      bodo.libs.array_kernels.setna(out_arr, i)\n'
        cgyx__glqje += '      continue\n'
        acnfd__ixnaq = 'lhs[i]' if myhqa__cscb else 'lhs'
        jdfs__xqhoc = 'rhs[i]' if hvg__iaa else 'rhs'
        cgyx__glqje += f'    out_arr[i] = op({acnfd__ixnaq}, {jdfs__xqhoc})\n'
        cgyx__glqje += '  return out_arr\n'
        cwn__ikd = {}
        exec(cgyx__glqje, {'bodo': bodo, 'numba': numba, 'op': op}, cwn__ikd)
        return cwn__ikd['impl']
    return overload_binary_cmp


lower_builtin('getiter', binary_array_type)(numba.np.arrayobj.getiter_array)


def pre_alloc_binary_arr_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


from numba.parfors.array_analysis import ArrayAnalysis
(ArrayAnalysis._analyze_op_call_bodo_libs_binary_arr_ext_pre_alloc_binary_array
    ) = pre_alloc_binary_arr_equiv

==> bodo-2022.4-cp38-cp38-macosx_10_15_x86_64.whl: bodo/libs/binary_arr_ext.py <==
"""Array implementation for binary (bytes) objects, which are usually immutable.
It is equivalent to string array, except that it stores a 'bytes' object for each
element instead of 'str'.
"""
import operator
import llvmlite.binding as ll
import numba
import numpy as np
from llvmlite import ir as lir
from numba.core import cgutils, types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import intrinsic, lower_builtin, lower_cast, make_attribute_wrapper, overload, overload_attribute, overload_method
import bodo
from bodo.libs import hstr_ext
from bodo.libs.array_item_arr_ext import ArrayItemArrayType
from bodo.utils.typing import BodoError, is_list_like_index_type
_bytes_fromhex = types.ExternalFunction('bytes_fromhex', types.int64(types.
    voidptr, types.voidptr, types.uint64))
ll.add_symbol('bytes_to_hex', hstr_ext.bytes_to_hex)
ll.add_symbol('bytes_fromhex', hstr_ext.bytes_fromhex)
bytes_type = types.Bytes(types.uint8, 1, 'C', readonly=True)


class BinaryArrayType(types.IterableType, types.ArrayCompatible):

    def __init__(self):
        super(BinaryArrayType, self).__init__(name='BinaryArrayType()')

    @property
    def as_array(self):
        return types.Array(types.undefined, 1, 'C')

    @property
    def dtype(self):
        return bytes_type

    def copy(self):
        return BinaryArrayType()

    @property
    def iterator_type(self):
        return bodo.utils.typing.BodoArrayIterator(self)


binary_array_type = BinaryArrayType()


@overload(len, no_unliteral=True)
def bin_arr_len_overload(bin_arr):
    if bin_arr == binary_array_type:
        return lambda bin_arr: len(bin_arr._data)


@overload_attribute(BinaryArrayType, 'size')
def bin_arr_size_overload(bin_arr):
    return lambda bin_arr: len(bin_arr._data)


@overload_attribute(BinaryArrayType, 'shape')
def bin_arr_shape_overload(bin_arr):
    return lambda bin_arr: (len(bin_arr._data),)


@overload_attribute(BinaryArrayType, 'nbytes')
def bin_arr_nbytes_overload(bin_arr):
    return lambda bin_arr: bin_arr._data.nbytes


@overload_attribute(BinaryArrayType, 'ndim')
def overload_bin_arr_ndim(A):
    return lambda A: 1


@overload_attribute(BinaryArrayType, 'dtype')
def overload_bool_arr_dtype(A):
    return lambda A: np.dtype('O')


@numba.njit
def pre_alloc_binary_array(n_bytestrs, n_chars):
    if n_chars is None:
        n_chars = -1
    bin_arr = init_binary_arr(bodo.libs.array_item_arr_ext.
        pre_alloc_array_item_array(np.int64(n_bytestrs), (np.int64(n_chars)
        ,), bodo.libs.str_arr_ext.char_arr_type))
    if n_chars == 0:
        bodo.libs.str_arr_ext.set_all_offsets_to_0(bin_arr)
    return bin_arr


@intrinsic
def init_binary_arr(typingctx, data_typ=None):
    assert isinstance(data_typ, ArrayItemArrayType
        ) and data_typ.dtype == types.Array(types.uint8, 1, 'C')

    def codegen(context, builder, sig, args):
        zzg__brlnc, = args
        dyi__eme = context.make_helper(builder, binary_array_type)
        dyi__eme.data = zzg__brlnc
        context.nrt.incref(builder, data_typ, zzg__brlnc)
        return dyi__eme._getvalue()
    return binary_array_type(data_typ), codegen


@intrinsic
def init_bytes_type(typingctx, data_typ, length_type):
    assert data_typ == types.Array(types.uint8, 1, 'C')
    assert length_type == types.int64

    def codegen(context, builder, sig, args):
        eyq__jovnr = cgutils.create_struct_proxy(sig.args[0])(context,
            builder, value=args[0])
        usj__xnggy = args[1]
        paw__lbjce = cgutils.create_struct_proxy(bytes_type)(context, builder)
        paw__lbjce.meminfo = context.nrt.meminfo_alloc(builder, usj__xnggy)
        paw__lbjce.nitems = usj__xnggy
        paw__lbjce.itemsize = lir.Constant(paw__lbjce.itemsize.type, 1)
        paw__lbjce.data = context.nrt.meminfo_data(builder, paw__lbjce.meminfo)
        paw__lbjce.parent = cgutils.get_null_value(paw__lbjce.parent.type)
        paw__lbjce.shape = cgutils.pack_array(builder, [usj__xnggy],
            context.get_value_type(types.intp))
        paw__lbjce.strides = eyq__jovnr.strides
        cgutils.memcpy(builder, paw__lbjce.data, eyq__jovnr.data, usj__xnggy)
        return paw__lbjce._getvalue()
    return bytes_type(data_typ, length_type), codegen


@intrinsic
def cast_bytes_uint8array(typingctx, data_typ):
    assert data_typ == bytes_type

    def codegen(context, builder, sig, args):
        return impl_ret_borrowed(context, builder, sig.return_type, args[0])
    return types.Array(types.uint8, 1, 'C')(data_typ), codegen


@overload_method(BinaryArrayType, 'copy', no_unliteral=True)
def binary_arr_copy_overload(arr):

    def copy_impl(arr):
        return init_binary_arr(arr._data.copy())
    return copy_impl


@overload_method(types.Bytes, 'hex')
def binary_arr_hex(arr):
    ieuxa__qkypc = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

    def impl(arr):
        usj__xnggy = len(arr) * 2
        output = numba.cpython.unicode._empty_string(ieuxa__qkypc,
            usj__xnggy, 1)
        bytes_to_hex(output, arr)
        return output
    return impl


@lower_cast(types.CPointer(types.uint8), types.voidptr)
def cast_uint8_array_to_voidptr(context, builder, fromty, toty, val):
    return val


make_attribute_wrapper(types.Bytes, 'data', '_data')


@overload_method(types.Bytes, '__hash__')
def bytes_hash(arr):

    def impl(arr):
        return numba.cpython.hashing._Py_HashBytes(arr._data, len(arr))
    return impl


@intrinsic
def bytes_to_hex(typingctx, output, arr):

    def codegen(context, builder, sig, args):
        ggut__ksftn = cgutils.create_struct_proxy(sig.args[0])(context,
            builder, value=args[0])
        mtf__laave = cgutils.create_struct_proxy(sig.args[1])(context,
            builder, value=args[1])
        qegik__ifurf = lir.FunctionType(lir.IntType(8).as_pointer(), [lir.
            IntType(8).as_pointer(), lir.IntType(8).as_pointer(), lir.
            IntType(64)])
        apz__jxt = cgutils.get_or_insert_function(builder.module,
            qegik__ifurf, name='bytes_to_hex')
        builder.call(apz__jxt, (ggut__ksftn.data, mtf__laave.data,
            mtf__laave.nitems))
    return types.void(output, arr), codegen


@overload(operator.getitem, no_unliteral=True)
def binary_arr_getitem(arr, ind):
    if arr != binary_array_type:
        return
    if isinstance(ind, types.Integer):

        def impl(arr, ind):
            ila__ogimr = arr._data[ind]
            return init_bytes_type(ila__ogimr, len(ila__ogimr))
        return impl
    if is_list_like_index_type(ind) and (ind.dtype == types.bool_ or
        isinstance(ind.dtype, types.Integer)) or isinstance(ind, types.
        SliceType):
        return lambda arr, ind: init_binary_arr(arr._data[ind])
    raise BodoError(
        f'getitem for Binary Array with indexing type {ind} not supported.')


def bytes_fromhex(hex_str):
    pass


@overload(bytes_fromhex)
def overload_bytes_fromhex(hex_str):
    hex_str = types.unliteral(hex_str)
    if hex_str == bodo.string_type:
        ieuxa__qkypc = numba.cpython.unicode.PY_UNICODE_1BYTE_KIND

        def impl(hex_str):
            if not hex_str._is_ascii or hex_str._kind != ieuxa__qkypc:
                raise TypeError(
                    'bytes.fromhex is only supported on ascii strings')
            zzg__brlnc = np.empty(len(hex_str) // 2, np.uint8)
            usj__xnggy = _bytes_fromhex(zzg__brlnc.ctypes, hex_str._data,
                len(hex_str))
            hdf__jfvv = init_bytes_type(zzg__brlnc, usj__xnggy)
            return hdf__jfvv
        return impl
    raise BodoError(f'bytes.fromhex not supported with argument type {hex_str}'
        )


@overload(operator.setitem)
def binary_arr_setitem(arr, ind, val):
    if arr != binary_array_type:
        return
    if val == types.none or isinstance(val, types.optional):
        return
    if val != bytes_type:
        raise BodoError(
            f'setitem for Binary Array only supported with bytes value and integer indexing'
            )
    if isinstance(ind, types.Integer):

        def impl(arr, ind, val):
            arr._data[ind] = bodo.libs.binary_arr_ext.cast_bytes_uint8array(val
                )
        return impl
    raise BodoError(
        f'setitem for Binary Array with indexing type {ind} not supported.')


def create_binary_cmp_op_overload(op):

    def overload_binary_cmp(lhs, rhs):
        htf__oxup = lhs == binary_array_type
        ueupg__bhph = rhs == binary_array_type
        rrup__tvhyf = 'lhs' if htf__oxup else 'rhs'
        cbprq__xki = 'def impl(lhs, rhs):\n'
        cbprq__xki += '  numba.parfors.parfor.init_prange()\n'
        cbprq__xki += f'  n = len({rrup__tvhyf})\n'
        cbprq__xki += (
            '  out_arr = bodo.libs.bool_arr_ext.alloc_bool_array(n)\n')
        cbprq__xki += '  for i in numba.parfors.parfor.internal_prange(n):\n'
        udgxp__nzb = []
        if htf__oxup:
            udgxp__nzb.append('bodo.libs.array_kernels.isna(lhs, i)')
        if ueupg__bhph:
            udgxp__nzb.append('bodo.libs.array_kernels.isna(rhs, i)')
        cbprq__xki += f"    if {' or '.join(udgxp__nzb)}:\n"
        cbprq__xki += '      bodo.libs.array_kernels.setna(out_arr, i)\n'
        cbprq__xki += '      continue\n'
        vzsem__krvby = 'lhs[i]' if htf__oxup else 'lhs'
        qfemy__abq = 'rhs[i]' if ueupg__bhph else 'rhs'
        cbprq__xki += f'    out_arr[i] = op({vzsem__krvby}, {qfemy__abq})\n'
        cbprq__xki += '  return out_arr\n'
        ivpzb__yvhd = {}
        exec(cbprq__xki, {'bodo': bodo, 'numba': numba, 'op': op}, ivpzb__yvhd)
        return ivpzb__yvhd['impl']
    return overload_binary_cmp


lower_builtin('getiter', binary_array_type)(numba.np.arrayobj.getiter_array)


def pre_alloc_binary_arr_equiv(self, scope, equiv_set, loc, args, kws):
    assert len(args) == 2 and not kws
    return ArrayAnalysis.AnalyzeResult(shape=args[0], pre=[])


from numba.parfors.array_analysis import ArrayAnalysis
(ArrayAnalysis._analyze_op_call_bodo_libs_binary_arr_ext_pre_alloc_binary_array
    ) = pre_alloc_binary_arr_equiv
